@@ -1423,11 +1423,12 @@ func TestCloseInConnectionStateCallback(t *testing.T) {
 
 	isClosed := make(chan interface{})
 	isConnected := make(chan interface{})
+	connectionStateConnectedSeen := make(chan interface{})
 	err = aAgent.OnConnectionStateChange(func(c ConnectionState) {
 		switch c {
 		case ConnectionStateConnected:
 			<-isConnected
-			assert.NoError(t, aAgent.Close())
+			close(connectionStateConnectedSeen)
 		case ConnectionStateClosed:
 			close(isClosed)
 		default:
@@ -1439,6 +1440,8 @@ func TestCloseInConnectionStateCallback(t *testing.T) {
 
 	connect(aAgent, bAgent)
 	close(isConnected)
+	<-connectionStateConnectedSeen
+	require.NoError(t, aAgent.Close())
 
 	<-isClosed
 	assert.NoError(t, bAgent.Close())
