@@ -19,6 +19,7 @@ import (
 	"runtime"
 	"strings"
 	"sync"
+	"sync/atomic"
 	"testing"
 	"time"
 
@@ -257,6 +258,7 @@ func createNewAssociationPair(br *test.Bridge, ackMode int, recvBufSize uint32)
 
 	go func() {
 		a0, err0 = Client(Config{
+			Name:                 "a0",
 			NetConn:              br.GetConn0(),
 			MaxReceiveBufferSize: recvBufSize,
 			LoggerFactory:        loggerFactory,
@@ -264,7 +266,11 @@ func createNewAssociationPair(br *test.Bridge, ackMode int, recvBufSize uint32)
 		handshake0Ch <- true
 	}()
 	go func() {
-		a1, err1 = Client(Config{
+		// we could have two "client"s here but it's more
+		// standard to have one peer starting initialization and
+		// another waiting for the initialization to be requested (INIT).
+		a1, err1 = Server(Config{
+			Name:                 "a1",
 			NetConn:              br.GetConn1(),
 			MaxReceiveBufferSize: recvBufSize,
 			LoggerFactory:        loggerFactory,
@@ -411,8 +417,8 @@ func establishSessionPair(br *test.Bridge, a0, a1 *Association, si uint16) (*Str
 }
 
 func TestAssocReliable(t *testing.T) {
-	// sbuf - small enogh not to be fragmented
-	//        large enobh not to be bundled
+	// sbuf - small enough not to be fragmented
+	//        large enough not to be bundled
 	sbuf := make([]byte, 1000)
 	for i := 0; i < len(sbuf); i++ {
 		sbuf[i] = byte(i & 0xff)
@@ -420,8 +426,8 @@ func TestAssocReliable(t *testing.T) {
 	rand.Seed(time.Now().UnixNano())
 	rand.Shuffle(len(sbuf), func(i, j int) { sbuf[i], sbuf[j] = sbuf[j], sbuf[i] })
 
-	// sbufL - large enogh to be fragmented into two chunks and each chunks are
-	//        large enobh not to be bundled
+	// sbufL - large enough to be fragmented into two chunks and each chunks are
+	//        large enough not to be bundled
 	sbufL := make([]byte, 2000)
 	for i := 0; i < len(sbufL); i++ {
 		sbufL[i] = byte(i & 0xff)
@@ -821,8 +827,8 @@ func TestAssocReliable(t *testing.T) {
 
 func TestAssocUnreliable(t *testing.T) {
 	// sbuf1, sbuf2:
-	//    large enogh to be fragmented into two chunks and each chunks are
-	//    large enobh not to be bundled
+	//    large enough to be fragmented into two chunks and each chunks are
+	//    large enough not to be bundled
 	sbuf1 := make([]byte, 2000)
 	sbuf2 := make([]byte, 2000)
 	for i := 0; i < len(sbuf1); i++ {
@@ -836,8 +842,8 @@ func TestAssocUnreliable(t *testing.T) {
 	rand.Seed(time.Now().UnixNano())
 	rand.Shuffle(len(sbuf2), func(i, j int) { sbuf2[i], sbuf2[j] = sbuf2[j], sbuf2[i] })
 
-	// sbuf - small enogh not to be fragmented
-	//        large enobh not to be bundled
+	// sbuf - small enough not to be fragmented
+	//        large enough not to be bundled
 	sbuf := make([]byte, 1000)
 	for i := 0; i < len(sbuf); i++ {
 		sbuf[i] = byte(i & 0xff)
@@ -1277,10 +1283,10 @@ func TestHandleForwardTSN(t *testing.T) {
 			LoggerFactory: loggerFactory,
 		})
 		a.useForwardTSN = true
-		prevTSN := a.peerLastTSN
+		prevTSN := a.peerLastTSN()
 
 		fwdtsn := &chunkForwardTSN{
-			newCumulativeTSN: a.peerLastTSN + 3,
+			newCumulativeTSN: prevTSN + 3,
 			streams:          []chunkForwardTSNStream{{identifier: 0, sequence: 0}},
 		}
 
@@ -1290,7 +1296,7 @@ func TestHandleForwardTSN(t *testing.T) {
 		delayedAckTriggered := a.delayedAckTriggered
 		immediateAckTriggered := a.immediateAckTriggered
 		a.lock.Unlock()
-		assert.Equal(t, a.peerLastTSN, prevTSN+3, "peerLastTSN should advance by 3 ")
+		assert.Equal(t, a.peerLastTSN(), prevTSN+3, "peerLastTSN should advance by 3 ")
 		assert.True(t, delayedAckTriggered, "delayed sack should be triggered")
 		assert.False(t, immediateAckTriggered, "immediate sack should NOT be triggered")
 		assert.Nil(t, p, "should return nil")
@@ -1302,20 +1308,13 @@ func TestHandleForwardTSN(t *testing.T) {
 			LoggerFactory: loggerFactory,
 		})
 		a.useForwardTSN = true
-		prevTSN := a.peerLastTSN
+		prevTSN := a.peerLastTSN()
 
 		// this chunk is blocked by the missing chunk at tsn=1
-		a.payloadQueue.push(&chunkPayloadData{
-			beginningFragment:    true,
-			endingFragment:       true,
-			tsn:                  a.peerLastTSN + 2,
-			streamIdentifier:     0,
-			streamSequenceNumber: 1,
-			userData:             []byte("ABC"),
-		}, a.peerLastTSN)
+		a.payloadQueue.push(a.peerLastTSN() + 2)
 
 		fwdtsn := &chunkForwardTSN{
-			newCumulativeTSN: a.peerLastTSN + 1,
+			newCumulativeTSN: a.peerLastTSN() + 1,
 			streams: []chunkForwardTSNStream{
 				{identifier: 0, sequence: 1},
 			},
@@ -1327,7 +1326,7 @@ func TestHandleForwardTSN(t *testing.T) {
 		delayedAckTriggered := a.delayedAckTriggered
 		immediateAckTriggered := a.immediateAckTriggered
 		a.lock.Unlock()
-		assert.Equal(t, a.peerLastTSN, prevTSN+2, "peerLastTSN should advance by 3")
+		assert.Equal(t, a.peerLastTSN(), prevTSN+2, "peerLastTSN should advance by 3")
 		assert.True(t, delayedAckTriggered, "delayed sack should be triggered")
 		assert.False(t, immediateAckTriggered, "immediate sack should NOT be triggered")
 		assert.Nil(t, p, "should return nil")
@@ -1339,20 +1338,13 @@ func TestHandleForwardTSN(t *testing.T) {
 			LoggerFactory: loggerFactory,
 		})
 		a.useForwardTSN = true
-		prevTSN := a.peerLastTSN
+		prevTSN := a.peerLastTSN()
 
 		// this chunk is blocked by the missing chunk at tsn=1
-		a.payloadQueue.push(&chunkPayloadData{
-			beginningFragment:    true,
-			endingFragment:       true,
-			tsn:                  a.peerLastTSN + 3,
-			streamIdentifier:     0,
-			streamSequenceNumber: 1,
-			userData:             []byte("ABC"),
-		}, a.peerLastTSN)
+		a.payloadQueue.push(a.peerLastTSN() + 3)
 
 		fwdtsn := &chunkForwardTSN{
-			newCumulativeTSN: a.peerLastTSN + 1,
+			newCumulativeTSN: a.peerLastTSN() + 1,
 			streams: []chunkForwardTSNStream{
 				{identifier: 0, sequence: 1},
 			},
@@ -1363,7 +1355,7 @@ func TestHandleForwardTSN(t *testing.T) {
 		a.lock.Lock()
 		immediateAckTriggered := a.immediateAckTriggered
 		a.lock.Unlock()
-		assert.Equal(t, a.peerLastTSN, prevTSN+1, "peerLastTSN should advance by 1")
+		assert.Equal(t, a.peerLastTSN(), prevTSN+1, "peerLastTSN should advance by 1")
 		assert.True(t, immediateAckTriggered, "immediate sack should be triggered")
 
 		assert.Nil(t, p, "should return nil")
@@ -1375,10 +1367,10 @@ func TestHandleForwardTSN(t *testing.T) {
 			LoggerFactory: loggerFactory,
 		})
 		a.useForwardTSN = true
-		prevTSN := a.peerLastTSN
+		prevTSN := a.peerLastTSN()
 
 		fwdtsn := &chunkForwardTSN{
-			newCumulativeTSN: a.peerLastTSN, // old TSN
+			newCumulativeTSN: a.peerLastTSN(), // old TSN
 			streams: []chunkForwardTSNStream{
 				{identifier: 0, sequence: 1},
 			},
@@ -1389,7 +1381,7 @@ func TestHandleForwardTSN(t *testing.T) {
 		a.lock.Lock()
 		ackState := a.ackState
 		a.lock.Unlock()
-		assert.Equal(t, a.peerLastTSN, prevTSN, "peerLastTSN should not advance")
+		assert.Equal(t, a.peerLastTSN(), prevTSN, "peerLastTSN should not advance")
 		assert.Equal(t, ackStateImmediate, ackState, "sack should be requested")
 		assert.Nil(t, p, "should return nil")
 	})
@@ -1622,7 +1614,7 @@ func TestAssocT1CookieTimer(t *testing.T) {
 		// Drop all COOKIE-ECHO
 		br.Filter(0, func(raw []byte) bool {
 			p := &packet{}
-			err := p.unmarshal(raw)
+			err := p.unmarshal(true, raw)
 			if !assert.Nil(t, err, "failed to parse packet") {
 				return false // drop
 			}
@@ -1684,7 +1676,7 @@ func TestAssocCreateNewStream(t *testing.T) {
 		toBeIgnored := &chunkPayloadData{
 			beginningFragment: true,
 			endingFragment:    true,
-			tsn:               a.peerLastTSN + 1,
+			tsn:               a.peerLastTSN() + 1,
 			streamIdentifier:  newSI,
 			userData:          []byte("ABC"),
 		}
@@ -1752,7 +1744,7 @@ func TestAssocT3RtxTimer(t *testing.T) {
 }
 
 func TestAssocCongestionControl(t *testing.T) {
-	// sbuf - large enobh not to be bundled
+	// sbuf - large enough not to be bundled
 	sbuf := make([]byte, 1000)
 	for i := 0; i < len(sbuf); i++ {
 		sbuf[i] = byte(i & 0xcc)
@@ -1823,7 +1815,7 @@ func TestAssocCongestionControl(t *testing.T) {
 		assert.False(t, inFastRecovery, "should not be in fast-recovery")
 
 		t.Logf("nDATAs      : %d\n", a1.stats.getNumDATAs())
-		t.Logf("nSACKs      : %d\n", a0.stats.getNumSACKs())
+		t.Logf("nSACKs      : %d\n", a0.stats.getNumSACKsReceived())
 		t.Logf("nAckTimeouts: %d\n", a1.stats.getNumAckTimeouts())
 		t.Logf("nFastRetrans: %d\n", a0.stats.getNumFastRetrans())
 
@@ -1907,11 +1899,11 @@ func TestAssocCongestionControl(t *testing.T) {
 		assert.Equal(t, 0, s1.getNumBytesInReassemblyQueue(), "reassembly queue should be empty")
 
 		t.Logf("nDATAs      : %d\n", a1.stats.getNumDATAs())
-		t.Logf("nSACKs      : %d\n", a0.stats.getNumSACKs())
+		t.Logf("nSACKs      : %d\n", a0.stats.getNumSACKsReceived())
 		t.Logf("nT3Timeouts : %d\n", a0.stats.getNumT3Timeouts())
 
 		assert.Equal(t, uint64(nPacketsToSend), a1.stats.getNumDATAs(), "packet count mismatch")
-		assert.True(t, a0.stats.getNumSACKs() <= nPacketsToSend/2, "too many sacks")
+		assert.True(t, a0.stats.getNumSACKsReceived() <= nPacketsToSend/2, "too many sacks")
 		assert.Equal(t, uint64(0), a0.stats.getNumT3Timeouts(), "should be no retransmit")
 
 		closeAssociationPair(br, a0, a1)
@@ -2002,7 +1994,7 @@ func TestAssocCongestionControl(t *testing.T) {
 		assert.Equal(t, 0, s1.getNumBytesInReassemblyQueue(), "reassembly queue should be empty")
 
 		t.Logf("nDATAs      : %d\n", a1.stats.getNumDATAs())
-		t.Logf("nSACKs      : %d\n", a0.stats.getNumSACKs())
+		t.Logf("nSACKs      : %d\n", a0.stats.getNumSACKsReceived())
 		t.Logf("nAckTimeouts: %d\n", a1.stats.getNumAckTimeouts())
 
 		closeAssociationPair(br, a0, a1)
@@ -2081,11 +2073,11 @@ func TestAssocDelayedAck(t *testing.T) {
 		assert.Equal(t, 0, s1.getNumBytesInReassemblyQueue(), "reassembly queue should be empty")
 
 		t.Logf("nDATAs      : %d\n", a1.stats.getNumDATAs())
-		t.Logf("nSACKs      : %d\n", a0.stats.getNumSACKs())
+		t.Logf("nSACKs      : %d\n", a0.stats.getNumSACKsReceived())
 		t.Logf("nAckTimeouts: %d\n", a1.stats.getNumAckTimeouts())
 
 		assert.Equal(t, uint64(1), a1.stats.getNumDATAs(), "DATA chunk count mismatch")
-		assert.Equal(t, a0.stats.getNumSACKs(), a1.stats.getNumDATAs(), "sack count should be equal to the number of data chunks")
+		assert.Equal(t, a0.stats.getNumSACKsReceived(), a1.stats.getNumDATAs(), "sack count should be equal to the number of data chunks")
 		assert.Equal(t, uint64(1), a1.stats.getNumAckTimeouts(), "ackTimeout count mismatch")
 		assert.Equal(t, uint64(0), a0.stats.getNumT3Timeouts(), "should be no retransmit")
 
@@ -2093,8 +2085,28 @@ func TestAssocDelayedAck(t *testing.T) {
 	})
 }
 
+func checkGoroutineLeaks(t *testing.T) {
+	// Get the count of goroutines at the start of the test.
+	initialGoroutines := runtime.NumGoroutine()
+	// Register a cleanup function to run after the test completes.
+	t.Cleanup(func() {
+		// Allow for up to 1 second for all goroutines to finish.
+		for i := 0; i < 10; i++ {
+			time.Sleep(100 * time.Millisecond)
+			if goroutines := runtime.NumGoroutine(); goroutines <= initialGoroutines {
+				return
+			}
+		}
+
+		// If we've gotten this far, not all goroutines have finished.
+		t.Errorf("leaked %d goroutines", runtime.NumGoroutine()-initialGoroutines)
+	})
+}
+
 func TestAssocReset(t *testing.T) {
 	t.Run("Close one way", func(t *testing.T) {
+		checkGoroutineLeaks(t)
+
 		lim := test.TimeOut(time.Second * 10)
 		defer lim.Stop()
 
@@ -2156,6 +2168,8 @@ func TestAssocReset(t *testing.T) {
 	})
 
 	t.Run("Close both ways", func(t *testing.T) {
+		checkGoroutineLeaks(t)
+
 		lim := test.TimeOut(time.Second * 10)
 		defer lim.Stop()
 
@@ -2173,6 +2187,7 @@ func TestAssocReset(t *testing.T) {
 
 		assert.Equal(t, 0, a0.bufferedAmount(), "incorrect bufferedAmount")
 
+		// send a message from s0 to s1
 		n, err := s0.WriteSCTP([]byte(msg), PayloadTypeWebRTCBinary)
 		if err != nil {
 			assert.FailNow(t, "failed due to earlier error")
@@ -2180,7 +2195,8 @@ func TestAssocReset(t *testing.T) {
 		assert.Equal(t, len(msg), n, "unexpected length of received data")
 		assert.Equal(t, len(msg), a0.bufferedAmount(), "incorrect bufferedAmount")
 
-		err = s0.Close() // send reset
+		// close s0 as soon as the message is sent
+		err = s0.Close()
 		if err != nil {
 			t.Error(err)
 		}
@@ -2213,7 +2229,8 @@ func TestAssocReset(t *testing.T) {
 			}
 		}
 
-		err = s1.Close() // send reset
+		// send reset from s1
+		err = s1.Close()
 		if err != nil {
 			t.Error(err)
 		}
@@ -2222,7 +2239,10 @@ func TestAssocReset(t *testing.T) {
 			for {
 				_, _, err = s0.ReadSCTP(buf)
 				assert.Equal(t, io.EOF, err, "should be EOF")
-				doneCh <- err
+				if err != nil {
+					doneCh <- err
+					return
+				}
 			}
 		}()
 
@@ -2257,7 +2277,7 @@ func TestAssocAbort(t *testing.T) {
 			errorCauseHeader: errorCauseHeader{code: protocolViolation},
 		}},
 	}
-	packet, err := a0.createPacket([]chunk{abort}).marshal()
+	packet, err := a0.marshalPacket(a0.createPacket([]chunk{abort}))
 	assert.NoError(t, err)
 
 	_, _, err = establishSessionPair(br, a0, a1, si)
@@ -2354,8 +2374,7 @@ func (c *fakeEchoConn) SetWriteDeadline(time.Time) error { return nil }
 func TestRoutineLeak(t *testing.T) {
 	loggerFactory := logging.NewDefaultLoggerFactory()
 	t.Run("Close failed", func(t *testing.T) {
-		runtime.GC()
-		n0 := runtime.NumGoroutine()
+		checkGoroutineLeaks(t)
 
 		conn := newFakeEchoConn(io.EOF)
 		a, err := Client(Config{NetConn: conn, LoggerFactory: loggerFactory})
@@ -2375,12 +2394,9 @@ func TestRoutineLeak(t *testing.T) {
 			t.Errorf("closeWriteLoopCh is expected to be closed, but not")
 		}
 		_ = a
-		runtime.GC()
-		assert.Equal(t, n0, runtime.NumGoroutine(), "goroutine is leaked")
 	})
 	t.Run("Connection closed by remote host", func(t *testing.T) {
-		runtime.GC()
-		n0 := runtime.NumGoroutine()
+		checkGoroutineLeaks(t)
 
 		conn := newFakeEchoConn(nil)
 		a, err := Client(Config{NetConn: conn, LoggerFactory: loggerFactory})
@@ -2401,8 +2417,6 @@ func TestRoutineLeak(t *testing.T) {
 		default:
 			t.Errorf("closeWriteLoopCh is expected to be closed, but not")
 		}
-		runtime.GC()
-		assert.Equal(t, n0, runtime.NumGoroutine(), "goroutine is leaked")
 	})
 }
 
@@ -2454,7 +2468,7 @@ func TestAssocHandleInit(t *testing.T) {
 			return
 		}
 		assert.NoError(t, err, "should succeed")
-		assert.Equal(t, init.initialTSN-1, a.peerLastTSN, "should match")
+		assert.Equal(t, init.initialTSN-1, a.peerLastTSN(), "should match")
 		assert.Equal(t, uint16(1001), a.myMaxNumOutboundStreams, "should match")
 		assert.Equal(t, uint16(1002), a.myMaxNumInboundStreams, "should match")
 		assert.Equal(t, uint32(5678), a.peerVerificationTag, "should match")
@@ -2542,77 +2556,388 @@ func TestAssocMaxMessageSize(t *testing.T) {
 	})
 }
 
-func createAssocs(t *testing.T) (a1, a2 *Association) {
-	addr1 := &net.UDPAddr{
-		IP:   net.IP{127, 0, 0, 1},
-		Port: 1234,
-	}
+type dumbConnInboundHandler func([]byte)
 
-	addr2 := &net.UDPAddr{
-		IP:   net.IP{127, 0, 0, 1},
-		Port: 5678,
-	}
+type dumbConn2 struct {
+	net.Conn
+	packets              [][]byte
+	closed               bool
+	localAddr            net.Addr
+	remoteAddr           net.Addr
+	remoteInboundHandler dumbConnInboundHandler
+	mutex                sync.Mutex
+	cond                 *sync.Cond
+}
 
-	udp1, err := net.DialUDP("udp", addr1, addr2)
-	if err != nil {
-		panic(err)
+func newDumbConn2(localAddr, remoteAddr net.Addr) *dumbConn2 {
+	c := &dumbConn2{
+		packets:    [][]byte{},
+		localAddr:  localAddr,
+		remoteAddr: remoteAddr,
 	}
+	c.cond = sync.NewCond(&c.mutex)
+	return c
+}
 
-	udp2, err := net.DialUDP("udp", addr2, addr1)
-	if err != nil {
-		panic(err)
+// Implement the net.Conn interface methods
+func (c *dumbConn2) Read(b []byte) (n int, err error) {
+	c.mutex.Lock()
+	defer c.mutex.Unlock()
+
+	for {
+		if len(c.packets) > 0 {
+			packet := c.packets[0]
+			c.packets = c.packets[1:]
+			n := copy(b, packet)
+			return n, nil
+		}
+
+		if c.closed {
+			return 0, io.EOF
+		}
+
+		c.cond.Wait()
 	}
+}
+
+func (c *dumbConn2) Write(b []byte) (int, error) {
+	c.mutex.Lock()
+	closed := c.closed
+	c.mutex.Unlock()
+
+	if closed {
+		return 0, &net.OpError{Op: "write", Net: "udp", Addr: c.remoteAddr, Err: net.ErrClosed}
+	}
+	c.remoteInboundHandler(b)
+	return len(b), nil
+}
+
+func (c *dumbConn2) Close() error {
+	c.mutex.Lock()
+	defer c.mutex.Unlock()
+
+	c.closed = true
+	c.cond.Signal()
+	return nil
+}
+
+func (c *dumbConn2) LocalAddr() net.Addr {
+	// Unused by Association
+	return c.localAddr
+}
+
+func (c *dumbConn2) RemoteAddr() net.Addr {
+	// Unused by Association
+	return c.remoteAddr
+}
+
+func (c *dumbConn2) inboundHandler(packet []byte) {
+	c.mutex.Lock()
+	defer c.mutex.Unlock()
+
+	if !c.closed {
+		c.packets = append(c.packets, packet)
+		c.cond.Signal()
+	}
+}
+
+// crateUDPConnPair creates a pair of net.UDPConn objects that are connected with each other
+func createUDPConnPair() (net.Conn, net.Conn) {
+	addr1 := &net.UDPAddr{IP: net.ParseIP("127.0.0.1"), Port: 1234}
+	addr2 := &net.UDPAddr{IP: net.ParseIP("127.0.0.1"), Port: 5678}
+	conn1 := newDumbConn2(addr1, addr2)
+	conn2 := newDumbConn2(addr2, addr1)
+	conn1.remoteInboundHandler = conn2.inboundHandler
+	conn2.remoteInboundHandler = conn1.inboundHandler
+	return conn1, conn2
+}
+
+func createAssocs() (*Association, *Association, error) {
+	udp1, udp2 := createUDPConnPair()
 
 	loggerFactory := logging.NewDefaultLoggerFactory()
 
-	a1Chan := make(chan *Association)
-	a2Chan := make(chan *Association)
+	a1Chan := make(chan interface{})
+	a2Chan := make(chan interface{})
+
+	ctx, cancel := context.WithTimeout(context.Background(), 2*time.Second)
+	defer cancel()
 
 	go func() {
-		a, err := Client(Config{
+		a, err2 := createClientWithContext(ctx, Config{
 			NetConn:       udp1,
 			LoggerFactory: loggerFactory,
 		})
-		require.NoError(t, err)
-
-		a1Chan <- a
+		if err2 != nil {
+			a1Chan <- err2
+		} else {
+			a1Chan <- a
+		}
 	}()
 
 	go func() {
-		a, err := Client(Config{
+		a, err2 := createClientWithContext(ctx, Config{
 			NetConn:       udp2,
 			LoggerFactory: loggerFactory,
 		})
-		require.NoError(t, err)
-
-		a2Chan <- a
+		if err2 != nil {
+			a2Chan <- err2
+		} else {
+			a2Chan <- a
+		}
 	}()
 
-	select {
-	case a1 = <-a1Chan:
-	case <-time.After(time.Second):
-		assert.Fail(t, "timed out waiting for a1")
-	}
+	var a1 *Association
+	var a2 *Association
 
-	select {
-	case a2 = <-a2Chan:
-	case <-time.After(time.Second):
-		assert.Fail(t, "timed out waiting for a2")
+loop:
+	for {
+		select {
+		case v1 := <-a1Chan:
+			switch v := v1.(type) {
+			case *Association:
+				a1 = v
+				if a2 != nil {
+					break loop
+				}
+			case error:
+				return nil, nil, v
+			}
+		case v2 := <-a2Chan:
+			switch v := v2.(type) {
+			case *Association:
+				a2 = v
+				if a1 != nil {
+					break loop
+				}
+			case error:
+				return nil, nil, v
+			}
+		}
 	}
+	return a1, a2, nil
+}
 
-	return a1, a2
+// udpDiscardReader blocks all reads after block is set to true.
+// This allows us to send arbitrary packets on a stream and block the packets received in response
+type udpDiscardReader struct {
+	net.Conn
+	ctx   context.Context
+	block atomic.Bool
+}
+
+func (d *udpDiscardReader) Read(b []byte) (n int, err error) {
+	if d.block.Load() {
+		<-d.ctx.Done()
+		return 0, d.ctx.Err()
+	}
+	return d.Conn.Read(b)
+}
+
+func createAssociationPair(udpConn1 net.Conn, udpConn2 net.Conn) (*Association, *Association, error) {
+	loggerFactory := logging.NewDefaultLoggerFactory()
+
+	a1Chan := make(chan interface{})
+	a2Chan := make(chan interface{})
+
+	ctx, cancel := context.WithTimeout(context.Background(), 2*time.Second)
+	defer cancel()
+
+	go func() {
+		a, err2 := createClientWithContext(ctx, Config{
+			NetConn:       udpConn1,
+			LoggerFactory: loggerFactory,
+		})
+		if err2 != nil {
+			a1Chan <- err2
+		} else {
+			a1Chan <- a
+		}
+	}()
+
+	go func() {
+		a, err2 := createClientWithContext(ctx, Config{
+			NetConn:              udpConn2,
+			LoggerFactory:        loggerFactory,
+			MaxReceiveBufferSize: 100_000,
+		})
+		if err2 != nil {
+			a2Chan <- err2
+		} else {
+			a2Chan <- a
+		}
+	}()
+
+	var a1 *Association
+	var a2 *Association
+
+loop:
+	for {
+		select {
+		case v1 := <-a1Chan:
+			switch v := v1.(type) {
+			case *Association:
+				a1 = v
+				if a2 != nil {
+					break loop
+				}
+			case error:
+				return nil, nil, v
+			}
+		case v2 := <-a2Chan:
+			switch v := v2.(type) {
+			case *Association:
+				a2 = v
+				if a1 != nil {
+					break loop
+				}
+			case error:
+				return nil, nil, v
+			}
+		}
+	}
+	return a1, a2, nil
+}
+
+func noErrorClose(t *testing.T, closeF func() error) {
+	t.Helper()
+	require.NoError(t, closeF())
+}
+
+// readMyNextTSN uses a lock to read the myNextTSN field of the association.
+// Avoids a data race.
+func readMyNextTSN(a *Association) uint32 {
+	a.lock.Lock()
+	defer a.lock.Unlock()
+	return a.myNextTSN
+}
+
+func TestAssociationReceiveWindow(t *testing.T) {
+	udp1, udp2 := createUDPConnPair()
+	ctx, cancel := context.WithCancel(context.Background())
+	dudp1 := &udpDiscardReader{Conn: udp1, ctx: ctx}
+	// a1 is the association used for sending data
+	// a2 is the association with receive window of 100kB which we will
+	// try to bypass
+	a1, a2, err := createAssociationPair(dudp1, udp2)
+	require.NoError(t, err)
+	defer noErrorClose(t, a2.Close)
+	defer noErrorClose(t, a1.Close)
+	s1, err := a1.OpenStream(1, PayloadTypeWebRTCBinary)
+	require.NoError(t, err)
+	defer s1.Close() // nolint:errcheck,gosec
+	_, err = s1.WriteSCTP([]byte("hello"), PayloadTypeWebRTCBinary)
+	require.NoError(t, err)
+	dudp1.block.Store(true)
+
+	_, err = s1.WriteSCTP([]byte("hello"), PayloadTypeWebRTCBinary)
+	require.NoError(t, err)
+	s2, err := a2.AcceptStream()
+	require.NoError(t, err)
+	require.Equal(t, uint16(1), s2.streamIdentifier)
+
+	done := make(chan bool)
+	go func() {
+		chunks := s1.packetize(make([]byte, 1000), PayloadTypeWebRTCBinary)
+		chunks = chunks[:1]
+		chunk := chunks[0]
+		// Fake the TSN and enqueue 1 chunk with a very high tsn in the payload queue
+		chunk.tsn = readMyNextTSN(a1) + 1e9
+		for chunk.tsn > readMyNextTSN(a1) {
+			select {
+			case <-done:
+				return
+			default:
+			}
+			chunk.tsn--
+			pp := a1.bundleDataChunksIntoPackets(chunks)
+			for _, p := range pp {
+				raw, err := p.marshal(true)
+				if err != nil {
+					return
+				}
+				_, err = a1.netConn.Write(raw)
+				if err != nil {
+					return
+				}
+			}
+			if chunk.tsn%10 == 0 {
+				time.Sleep(10 * time.Millisecond)
+			}
+		}
+	}()
+
+	for cnt := 0; cnt < 15; cnt++ {
+		bytesQueued := s2.getNumBytesInReassemblyQueue()
+		if bytesQueued > 5_000_000 {
+			t.Error("too many bytes enqueued with receive window of 10kb", bytesQueued)
+			break
+		}
+		t.Log("bytes queued", bytesQueued)
+		time.Sleep(1 * time.Second)
+	}
+	close(done)
+	cancel()
+}
+
+func TestAssociationMaxTSNOffset(t *testing.T) {
+	udp1, udp2 := createUDPConnPair()
+	// a1 is the association used for sending data
+	// a2 is the association with receive window of 100kB which we will
+	// try to bypass
+	a1, a2, err := createAssociationPair(udp1, udp2)
+	require.NoError(t, err)
+	defer noErrorClose(t, a2.Close)
+	defer noErrorClose(t, a1.Close)
+	s1, err := a1.OpenStream(1, PayloadTypeWebRTCBinary)
+	require.NoError(t, err)
+	defer noErrorClose(t, s1.Close)
+	_, err = s1.WriteSCTP([]byte("hello"), PayloadTypeWebRTCBinary)
+	require.NoError(t, err)
+	_, err = s1.WriteSCTP([]byte("hello"), PayloadTypeWebRTCBinary)
+	require.NoError(t, err)
+	s2, err := a2.AcceptStream()
+	require.NoError(t, err)
+	require.Equal(t, uint16(1), s2.streamIdentifier)
+
+	chunks := s1.packetize(make([]byte, 1000), PayloadTypeWebRTCBinary)
+	chunks = chunks[:1]
+	sendChunk := func(tsn uint32) {
+		chunk := chunks[0]
+		// Fake the TSN and enqueue 1 chunk with a very high tsn in the payload queue
+		chunk.tsn = tsn
+		pp := a1.bundleDataChunksIntoPackets(chunks)
+		for _, p := range pp {
+			raw, err := p.marshal(true)
+			if err != nil {
+				t.Fatal(err)
+				return
+			}
+			_, err = a1.netConn.Write(raw)
+			if err != nil {
+				t.Fatal(err)
+				return
+			}
+		}
+	}
+	sendChunk(readMyNextTSN(a1) + 100_000)
+	time.Sleep(100 * time.Millisecond)
+	require.Less(t, s2.getNumBytesInReassemblyQueue(), 1000)
+
+	sendChunk(readMyNextTSN(a1) + 10_000)
+	time.Sleep(100 * time.Millisecond)
+	require.Less(t, s2.getNumBytesInReassemblyQueue(), 1000)
+
+	sendChunk(readMyNextTSN(a1) + minTSNOffset - 100)
+	time.Sleep(100 * time.Millisecond)
+	require.Greater(t, s2.getNumBytesInReassemblyQueue(), 1000)
 }
 
 func TestAssociation_Shutdown(t *testing.T) {
-	runtime.GC()
-	n0 := runtime.NumGoroutine()
+	checkGoroutineLeaks(t)
 
-	defer func() {
-		runtime.GC()
-		assert.Equal(t, n0, runtime.NumGoroutine(), "goroutine is leaked")
-	}()
-
-	a1, a2 := createAssocs(t)
+	a1, a2, err := createAssocs()
+	require.NoError(t, err)
 
 	s11, err := a1.OpenStream(1, PayloadTypeWebRTCString)
 	require.NoError(t, err)
@@ -2647,15 +2972,10 @@ func TestAssociation_Shutdown(t *testing.T) {
 }
 
 func TestAssociation_ShutdownDuringWrite(t *testing.T) {
-	runtime.GC()
-	n0 := runtime.NumGoroutine()
+	checkGoroutineLeaks(t)
 
-	defer func() {
-		runtime.GC()
-		assert.Equal(t, n0, runtime.NumGoroutine(), "goroutine is leaked")
-	}()
-
-	a1, a2 := createAssocs(t)
+	a1, a2, err := createAssocs()
+	require.NoError(t, err)
 
 	s11, err := a1.OpenStream(1, PayloadTypeWebRTCString)
 	require.NoError(t, err)
@@ -2851,7 +3171,7 @@ func TestAssociation_HandlePacketInCookieWaitState(t *testing.T) {
 				}()
 			}
 
-			packet, err := testCase.inputPacket.marshal()
+			packet, err := a.marshalPacket(testCase.inputPacket)
 			assert.NoError(t, err)
 			_, err = charlieConn.Write(packet)
 			assert.NoError(t, err)
@@ -2863,15 +3183,10 @@ func TestAssociation_HandlePacketInCookieWaitState(t *testing.T) {
 }
 
 func TestAssociation_Abort(t *testing.T) {
-	runtime.GC()
-	n0 := runtime.NumGoroutine()
+	checkGoroutineLeaks(t)
 
-	defer func() {
-		runtime.GC()
-		assert.Equal(t, n0, runtime.NumGoroutine(), "goroutine is leaked")
-	}()
-
-	a1, a2 := createAssocs(t)
+	a1, a2, err := createAssocs()
+	require.NoError(t, err)
 
 	s11, err := a1.OpenStream(1, PayloadTypeWebRTCString)
 	require.NoError(t, err)
@@ -2904,3 +3219,239 @@ func TestAssociation_Abort(t *testing.T) {
 	assert.Equal(t, i, 0, "expected no data read")
 	assert.Error(t, err, "User Initiated Abort: 1234", "expected abort reason")
 }
+
+// TestAssociation_createClientWithContext tests that the client is closed when the context is canceled.
+func TestAssociation_createClientWithContext(t *testing.T) {
+	checkGoroutineLeaks(t)
+
+	udp1, udp2 := createUDPConnPair()
+
+	loggerFactory := logging.NewDefaultLoggerFactory()
+
+	errCh1 := make(chan error)
+	errCh2 := make(chan error)
+
+	ctx, cancel := context.WithTimeout(context.Background(), 2*time.Second)
+
+	go func() {
+		_, err2 := createClientWithContext(ctx, Config{
+			NetConn:       udp1,
+			LoggerFactory: loggerFactory,
+		})
+		if err2 != nil {
+			errCh1 <- err2
+		} else {
+			errCh1 <- nil
+		}
+	}()
+
+	go func() {
+		_, err2 := createClientWithContext(ctx, Config{
+			NetConn:       udp2,
+			LoggerFactory: loggerFactory,
+		})
+		if err2 != nil {
+			errCh2 <- err2
+		} else {
+			errCh2 <- nil
+		}
+	}()
+
+	// Cancel the context immediately
+	cancel()
+
+	var err1 error
+	var err2 error
+loop:
+	for {
+		select {
+		case err1 = <-errCh1:
+			if err1 != nil && err2 != nil {
+				break loop
+			}
+		case err2 = <-errCh2:
+			if err1 != nil && err2 != nil {
+				break loop
+			}
+		}
+	}
+
+	assert.Error(t, err1, "context canceled")
+	assert.Error(t, err2, "context canceled")
+}
+
+type customLogger struct {
+	expectZeroChecksum bool
+	t                  *testing.T
+}
+
+func (c customLogger) Trace(string)                  {}
+func (c customLogger) Tracef(string, ...interface{}) {}
+func (c customLogger) Debug(string)                  {}
+func (c customLogger) Debugf(format string, args ...interface{}) {
+	if format == "[%s] sendZeroChecksum=%t (on initAck)" {
+		assert.Equal(c.t, args[1], c.expectZeroChecksum)
+	}
+}
+func (c customLogger) Info(string)                   {}
+func (c customLogger) Infof(string, ...interface{})  {}
+func (c customLogger) Warn(string)                   {}
+func (c customLogger) Warnf(string, ...interface{})  {}
+func (c customLogger) Error(string)                  {}
+func (c customLogger) Errorf(string, ...interface{}) {}
+
+func (c customLogger) NewLogger(string) logging.LeveledLogger {
+	return c
+}
+
+func TestAssociation_ZeroChecksum(t *testing.T) {
+	checkGoroutineLeaks(t)
+
+	lim := test.TimeOut(time.Second * 10)
+	defer lim.Stop()
+
+	for _, testCase := range []struct {
+		clientZeroChecksum, serverZeroChecksum, expectChecksumEnabled bool
+	}{
+		{true, true, true},
+		{false, false, false},
+		{true, false, false},
+		{false, true, true},
+	} {
+		a1chan, a2chan := make(chan *Association), make(chan *Association)
+
+		udp1, udp2 := createUDPConnPair()
+
+		go func() {
+			a1, err := Client(Config{
+				NetConn:            udp1,
+				LoggerFactory:      &customLogger{testCase.expectChecksumEnabled, t},
+				EnableZeroChecksum: testCase.clientZeroChecksum,
+			})
+			assert.NoError(t, err)
+			a1chan <- a1
+		}()
+
+		go func() {
+			a2, err := Server(Config{
+				NetConn:            udp2,
+				LoggerFactory:      &customLogger{testCase.expectChecksumEnabled, t},
+				EnableZeroChecksum: testCase.serverZeroChecksum,
+			})
+			assert.NoError(t, err)
+			a2chan <- a2
+		}()
+
+		a1, a2 := <-a1chan, <-a2chan
+
+		writeStream, err := a1.OpenStream(1, PayloadTypeWebRTCString)
+		require.NoError(t, err)
+
+		readStream, err := a2.OpenStream(1, PayloadTypeWebRTCString)
+		require.NoError(t, err)
+
+		testData := []byte("test")
+		_, err = writeStream.Write(testData)
+		require.NoError(t, err)
+
+		buf := make([]byte, len(testData))
+		_, err = readStream.Read(buf)
+		assert.NoError(t, err)
+		assert.Equal(t, testData, buf)
+
+		require.NoError(t, a1.Close())
+		require.NoError(t, a2.Close())
+	}
+}
+
+func TestDataChunkBundlingIntoPacket(t *testing.T) {
+	a := &Association{mtu: initialMTU}
+	chunks := make([]*chunkPayloadData, 300)
+	for i := 0; i < 300; i++ {
+		chunks[i] = &chunkPayloadData{userData: []byte{1}}
+	}
+	packets := a.bundleDataChunksIntoPackets(chunks)
+	for _, p := range packets {
+		raw, err := p.marshal(false)
+		require.NoError(t, err)
+		if len(raw) > int(initialMTU) {
+			t.Error("packet too long: ", len(raw))
+		}
+	}
+}
+
+func TestAssociation_ReconfigRequestsLimited(t *testing.T) {
+	checkGoroutineLeaks(t)
+
+	lim := test.TimeOut(time.Second * 10)
+	defer lim.Stop()
+
+	a1chan, a2chan := make(chan *Association), make(chan *Association)
+
+	udp1, udp2 := createUDPConnPair()
+
+	go func() {
+		a1, err := Client(Config{
+			NetConn:       udp1,
+			LoggerFactory: logging.NewDefaultLoggerFactory(),
+		})
+		assert.NoError(t, err)
+		a1chan <- a1
+	}()
+
+	go func() {
+		a2, err := Server(Config{
+			NetConn:       udp2,
+			LoggerFactory: logging.NewDefaultLoggerFactory(),
+		})
+		assert.NoError(t, err)
+		a2chan <- a2
+	}()
+
+	a1, a2 := <-a1chan, <-a2chan
+
+	writeStream, err := a1.OpenStream(1, PayloadTypeWebRTCString)
+	require.NoError(t, err)
+
+	readStream, err := a2.OpenStream(1, PayloadTypeWebRTCString)
+	require.NoError(t, err)
+
+	// exchange some data
+	testData := []byte("test")
+	_, err = writeStream.Write(testData)
+	require.NoError(t, err)
+
+	buf := make([]byte, len(testData))
+	_, err = readStream.Read(buf)
+	assert.NoError(t, err)
+	assert.Equal(t, testData, buf)
+
+	a1.lock.RLock()
+	tsn := a1.myNextTSN
+	a1.lock.RUnlock()
+	for i := 0; i < maxReconfigRequests+100; i++ {
+		c := &chunkReconfig{
+			paramA: &paramOutgoingResetRequest{
+				reconfigRequestSequenceNumber: 10 + uint32(i),
+				senderLastTSN:                 tsn + 10, // has to be enqueued
+				streamIdentifiers:             []uint16{uint16(i)},
+			},
+		}
+		p := a1.createPacket([]chunk{c})
+		buf, err := p.marshal(true)
+		require.NoError(t, err)
+		_, err = a1.netConn.Write(buf)
+		require.NoError(t, err)
+		if i%100 == 0 {
+			time.Sleep(100 * time.Millisecond)
+		}
+	}
+	// Let a2 process the requests
+	time.Sleep(2 * time.Second)
+	a2.lock.RLock()
+	require.LessOrEqual(t, len(a2.reconfigRequests), maxReconfigRequests)
+	a2.lock.RUnlock()
+
+	require.NoError(t, a1.Close())
+	require.NoError(t, a2.Close())
+}
